/******************************************************************************
 * Copyright (C) 2018  Tadeus Prastowo <eus@member.fsf.org>                   *
 *                                            (GPG public key ID: 0x277B48A6) *
 *                                                                            *
 * This program is free software: you can redistribute it and/or modify       *
 * it under the terms of the GNU General Public License as published by       *
 * the Free Software Foundation, either version 3 of the License, or          *
 * (at your option) any later version.                                        *
 *                                                                            *
 * This program is distributed in the hope that it will be useful,            *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              *
 * GNU General Public License for more details.                               *
 *                                                                            *
 * You should have received a copy of the GNU General Public License          *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.      *
 ******************************************************************************/

namespace tice {
  namespace v1 {
    namespace internals {
      namespace comp {
        
        template<typename... args>
        struct check_arg_1 : std::false_type {
        };

        template<std::intmax_t... ratio_args, typename... args>
        struct check_arg_1<std::ratio<ratio_args...>, args...> : std::true_type {
        };

      };
      namespace comp {

        template<typename... args>
        struct check_arg_2 : std::false_type {
        };

        template<typename arg__1, typename R, typename... Ts, R (*arg__function_ptr)(Ts...), typename... args>
        struct check_arg_2<arg__1, std::integral_constant<R(*)(Ts...), arg__function_ptr>, args...> : std::true_type {
        };
        
      };
      namespace comp {

        template<typename arg__ratio__wcet, typename arg__integral_constant__function_ptr>
        class Comp {
          static_assert(check_arg_1<arg__ratio__wcet, arg__integral_constant__function_ptr>::value,
                        "WCET is not specified using class template tice::v1::Ratio");
          static_assert(check_arg_2<arg__ratio__wcet, arg__integral_constant__function_ptr>::value,
                        "Computation is not passed as a function pointer");
          static_assert(std::ratio_greater<arg__ratio__wcet, std::ratio<0>>::value,
                        "WCET is not positive");
        public:
          typedef arg__ratio__wcet wcet;
          typedef utility::fn_info<arg__integral_constant__function_ptr> fn_info;
        };

      };
    };
  };
};
