/******************************************************************************
 * Copyright (C) 2018  Tadeus Prastowo <eus@member.fsf.org>                   *
 *                                            (GPG public key ID: 0x277B48A6) *
 *                                                                            *
 * This program is free software: you can redistribute it and/or modify       *
 * it under the terms of the GNU General Public License as published by       *
 * the Free Software Foundation, either version 3 of the License, or          *
 * (at your option) any later version.                                        *
 *                                                                            *
 * This program is distributed in the hope that it will be useful,            *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              *
 * GNU General Public License for more details.                               *
 *                                                                            *
 * You should have received a copy of the GNU General Public License          *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.      *
 ******************************************************************************/

namespace tice {
  namespace v1 {
    namespace internals {
      namespace feeder {

        template<typename... args>
        struct check_arg_1 : std::false_type {
        };

        template<typename... node_args, typename... args>
        struct check_arg_1<Node<node_args...>, args...> : std::true_type {
        };
        
      };
      namespace feeder {

        template<int arg__n, typename... args>
        struct check_args_pair : std::false_type {
          static_assert(arg__n == -1 /* this must always be false */,
                        "Channel and producer are not specified using class template tice::v1::Pair");
        };

        template<int arg__n, typename... pair_args>
        struct check_args_pair<arg__n, std::pair<pair_args...>> : std::true_type {
        };

        template<int arg__n, typename... pair_args, typename... args>
        struct check_args_pair<arg__n, std::pair<pair_args...>, args...>
          : check_args_pair<arg__n + 1, args...> {
        };
        
      };
      namespace feeder {

        template<int arg__n, typename... args>
        struct check_args_pair_fst : std::false_type {
          static_assert(arg__n == -1 /* this must always be false */,
                        "Channel type and initial value are not specified using class template tice::v1::Chan_inline_init or tice::v1::Chan");
        };

#define generate(chan_t, init_val_t)                                                            \
        template<int arg__n, typename T, init_val_t v, typename arg__2>                         \
        struct check_args_pair_fst<arg__n, std::pair<chan_t<T, v>, arg__2>> : std::true_type {  \
        };                                                                                      \
        template<int arg__n, typename T, init_val_t v, typename arg__2, typename... args>       \
        struct check_args_pair_fst<arg__n, std::pair<chan_t<T, v>, arg__2>, args...>            \
          : check_args_pair_fst<arg__n + 1, args...> {                                          \
        };

        generate(Chan_inline_init, T)
        generate(Chan, std::add_pointer_t<std::add_const_t<T>>)

#undef generate
        
      };
      namespace feeder {

        template<int arg__n, typename... args>
        struct check_args_pair_snd : std::false_type {
          static_assert(arg__n == -1 /* this must always be false */,
                        "Producer is not specified using class template tice::v1::Node");
        };

        template<int arg__n, typename arg__1, typename... node_args>
        struct check_args_pair_snd<arg__n, std::pair<arg__1, Node<node_args...>>> : std::true_type {
        };

        template<int arg__n, typename arg__1, typename... node_args, typename... args>
        struct check_args_pair_snd<arg__n, std::pair<arg__1, Node<node_args...>>, args...>
          : check_args_pair_snd<arg__n + 1, args...> {
        };
        
      };
      namespace feeder {

        template<bool equal, int nth_arg, int mth_arg>
        struct ensure_distinct_functions_pairwise_result : std::true_type {
        };

        template<int nth_arg, int mth_arg>
        struct ensure_distinct_functions_pairwise_result<false, nth_arg, mth_arg> : std::false_type {
          static_assert(nth_arg == -1 /* this must always be false */,
                        "Duplicated nodes");
        };

        template<int arg__head_is_nth_arg,
                 int arg__next_is_mth_arg,
                 typename arg__Node__head,
                 typename... args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer>
        struct ensure_distinct_functions_pairwise;

        template<int n, int m, typename... head_args>
        struct ensure_distinct_functions_pairwise<n, m, Node<head_args...>> : std::true_type {
        };

        template<int n, int m, typename... head_args, typename next_fst, typename... next_args>
        struct ensure_distinct_functions_pairwise<n, m, Node<head_args...>, std::pair<next_fst, Node<next_args...>>>
          : ensure_distinct_functions_pairwise_result<!utility::nontype_nontemplate_args_eq<decltype(&Node<head_args...>::fn_info::fn::value),
                                                                                            decltype(&Node<next_args...>::fn_info::fn::value),
                                                                                            &Node<head_args...>::fn_info::fn::value,
                                                                                            &Node<next_args...>::fn_info::fn::value>::value, n, m> {
        };

        template<int n, int m, typename... head_args, typename next, typename... rest>
        struct ensure_distinct_functions_pairwise<n, m, Node<head_args...>, next, rest...>
          : std::integral_constant<bool,
                                   ensure_distinct_functions_pairwise<n, m, Node<head_args...>, next>::value
                                   && ensure_distinct_functions_pairwise<n, m + 1, Node<head_args...>, rest...>::value> {
        };

        template<int arg__head_is_nth_arg,
                 typename arg__Node__head,
                 typename... args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer>
        struct ensure_distinct_functions;

        template<int n, typename... head_args>
        struct ensure_distinct_functions<n, Node<head_args...>> : std::true_type {
        };

        template<int n, typename... head_args, typename tail_fst, typename... tail_args, typename... rest>
        struct ensure_distinct_functions<n, Node<head_args...>, std::pair<tail_fst, Node<tail_args...>>, rest...>
          : std::integral_constant<bool,
                                   ensure_distinct_functions_pairwise<n, n + 1, Node<head_args...>, std::pair<tail_fst, Node<tail_args...>>,
                                                                      rest...>::value
                                   && ensure_distinct_functions<n + 1, Node<tail_args...>, rest...>::value> {
        };
        
      };
      namespace feeder {

        template<int arg__init_val_is_nth_arg,
                 typename... args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer>
        struct ensure_init_val_initializes_channel;

        template<bool arg__init_val_initializes_channel,
                 int arg__init_val_is_nth_arg,
                 typename... args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer>
        struct ensure_init_val_initializes_channel_result : std::false_type {
          static_assert(arg__init_val_is_nth_arg == -1 /* this must always be false */,
                        "Initial value cannot initialize channel due to type mismatch");
        };

        template<int n, typename... rest>
        struct ensure_init_val_initializes_channel_result<true, n, rest...>
          : ensure_init_val_initializes_channel<n + 1, rest...> {
        };

        template<int n>
        struct ensure_init_val_initializes_channel<n> : std::true_type {
        };

#define generate(chan_t, init_val_t, init_val)                                                                                                  \
        template<int n, typename T, init_val_t v, typename producer, typename... rest>                                                          \
        struct ensure_init_val_initializes_channel<n, std::pair<chan_t<T, v>, producer>, rest...>                                               \
          : ensure_init_val_initializes_channel_result<std::is_constructible<std::remove_cv_t<T>, decltype(init_val)>::value, n, rest...> {     \
        };

        generate(Chan_inline_init, T, v)
        generate(Chan, std::add_pointer_t<std::add_const_t<T>>, *v)

#undef generate
        
      };
      namespace feeder {

        template<int arg__producer_is_nth_arg,
                 typename... args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer>
        struct ensure_produced_data_assignable_to_channel;

        template<bool arg__produced_data_assignable_to_channel,
                 int arg__producer_is_nth_arg,
                 typename... args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer>
        struct ensure_produced_data_assignable_to_channel_result : std::false_type {
          static_assert(arg__producer_is_nth_arg == -1 /* this must always be false */,
                        "Producer's output is not assignable to channel");
        };

        template<int n, typename... rest>
        struct ensure_produced_data_assignable_to_channel_result<true, n, rest...>
          : ensure_produced_data_assignable_to_channel<n + 1, rest...> {
        };

        template<int n>
        struct ensure_produced_data_assignable_to_channel<n> : std::true_type {
        };

#define generate(chan_t, init_val_t)                                                                                                            \
        template<int n, typename T, init_val_t v, typename producer, typename... rest>                                                          \
        struct ensure_produced_data_assignable_to_channel<n, std::pair<chan_t<T, v>, producer>, rest...>                                        \
          : ensure_produced_data_assignable_to_channel_result<utility::simple_assignment_ok<std::remove_cv_t<T>,                                \
                                                                                            typename producer::fn_info::return_type>::value,    \
                                                              n, rest...> {                                                                     \
        };

        generate(Chan_inline_init, T)
        generate(Chan, std::add_pointer_t<std::add_const_t<T>>)

#undef generate
        
      };
      namespace feeder {
        
        template<typename arg__Node__consumer,
                 int arg__channel_is_nth_arg,
                 typename arg__tuple__of__channel_types,
                 typename... args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer>
        struct ensure_channel_initializes_consumer;

        template<bool arg__channel_initializes_consumer,
                 int arg__channel_is_nth_arg,
                 typename arg__Node__consumer,
                 typename arg__tuple__of__channel_types,
                 typename... args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer>
        struct ensure_channel_initializes_consumer_result : std::false_type {
          static_assert(arg__channel_is_nth_arg == -1 /* this must always be false */,
                        "Channel cannot initialize consumer's corresponding parameter");
        };

        template<typename consumer, int n, typename channel_type_accumulator>
        struct ensure_channel_initializes_consumer_result<true, n, consumer, channel_type_accumulator>
          : utility::is_callable<typename consumer::fn_info::fn::value_type, channel_type_accumulator> {
        };

        template<typename consumer, int n, typename channel_type_accumulator, typename... rest>
        struct ensure_channel_initializes_consumer_result<true, n, consumer, channel_type_accumulator, rest...>
          : ensure_channel_initializes_consumer<consumer, n + 1, channel_type_accumulator, rest...> {
        };

#define generate(chan_t, init_val_t)                                                                                                               \
        template<typename consumer, int n, typename... channel_types, typename T, init_val_t v, typename producer, typename... rest>               \
        struct ensure_channel_initializes_consumer<consumer, n, std::tuple<channel_types...>, std::pair<chan_t<T, v>, producer>, rest...>          \
          : ensure_channel_initializes_consumer_result<std::is_constructible<std::tuple_element_t<n - 2, typename consumer::fn_info::param_types>, \
                                                                             std::add_const_t<std::remove_cv_t<T>>>::value,                        \
                                                       n,                                                                                          \
                                                       consumer,                                                                                   \
                                                       std::tuple<channel_types..., std::add_const_t<std::remove_cv_t<T>>>,                        \
                                                       rest...> {                                                                                  \
        };
        
        generate(Chan_inline_init, T)
        generate(Chan, std::add_pointer_t<std::add_const_t<T>>)

#undef generate
        
      };
    };
  };
};

namespace tice {
  namespace v1 {

    template<typename arg__Node__consumer,
             typename arg__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer,
             typename... args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer>
    class Feeder {
      static_assert(internals::feeder::check_arg_1<
                    arg__Node__consumer,
                    arg__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer,
                    args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer...
                    >::value,
                    "Consumer is not specified using class template tice::v1::Node");
      static_assert(std::tuple_size<typename arg__Node__consumer::fn_info::param_types>::value
                    == 1 + sizeof...(args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer),
                    "Consumer's parameter count and incoming arc count differ");
      static_assert(internals::feeder::check_args_pair<
                    2,
                    arg__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer,
                    args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer...
                    >::value
                    && internals::feeder::check_args_pair_fst<
                    2,
                    arg__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer,
                    args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer...
                    >::value
                    && internals::feeder::check_args_pair_snd<
                    2,
                    arg__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer,
                    args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer...
                    >::value
                    && internals::feeder::ensure_distinct_functions<
                    1,
                    arg__Node__consumer,
                    arg__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer,
                    args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer...
                    >::value
                    && internals::feeder::ensure_init_val_initializes_channel<
                    2,
                    arg__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer,
                    args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer...
                    >::value
                    && internals::feeder::ensure_produced_data_assignable_to_channel<
                    2,
                    arg__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer,
                    args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer...
                    >::value,
                    "Error involving argument 2 and beyond");
      static_assert(internals::feeder::ensure_channel_initializes_consumer<
                    arg__Node__consumer,
                    2,
                    std::tuple<>,
                    arg__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer,
                    args__pair__of__Chan_inline_init_or_Chan__channel__and__Node__producer...
                    >::value,
                    "Some channels and consumer's parameters are incompatible");
    };
    
  };
};
