/******************************************************************************
 * Copyright (C) 2018  Tadeus Prastowo <eus@member.fsf.org>                   *
 *                                            (GPG public key ID: 0x277B48A6) *
 *                                                                            *
 * This program is free software: you can redistribute it and/or modify       *
 * it under the terms of the GNU General Public License as published by       *
 * the Free Software Foundation, either version 3 of the License, or          *
 * (at your option) any later version.                                        *
 *                                                                            *
 * This program is distributed in the hope that it will be useful,            *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              *
 * GNU General Public License for more details.                               *
 *                                                                            *
 * You should have received a copy of the GNU General Public License          *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.      *
 ******************************************************************************/

namespace tice {
  namespace v1 {
    namespace internals {
      namespace utility {

	template<typename T1, typename T2, T1 arg1, T2 arg2>
        struct nontype_nontemplate_args_eq : std::false_type {
        };
        template<typename T, T arg>
        struct nontype_nontemplate_args_eq<T, T, arg, arg> : std::true_type {
        };

      };
      namespace utility {

        template<typename arg__integral_constant__function_ptr>
        struct fn_info;

        template<typename R, typename... Ts, R (*arg__function_ptr)(Ts...)>
        struct fn_info<std::integral_constant<R(*)(Ts...), arg__function_ptr>> {
          typedef R return_type;
          typedef std::tuple<Ts...> param_types;
          typedef std::integral_constant<R(*)(Ts...), arg__function_ptr> fn;
        };
        
      };
      namespace utility {

        template<typename LHS, typename RHS>
        struct simple_assignment_ok /* based on ISO-IEC 14882:2014 (C++14 standard) [expr.ass]p{3,4} */
          : std::integral_constant<bool,
                                   std::is_union<LHS>::value || std::is_class<LHS>::value
                                   ? std::is_assignable<LHS, RHS>::value
                                   : std::is_convertible<RHS, std::remove_cv_t<LHS>>::value> {
        };
        
      };
      namespace utility {

        struct Y {
        };
        
        struct N {
        };
        
        template<typename T>
        constexpr Y has_member_typedef_type(typename T::type *);
        
        template<typename T>
        constexpr N has_member_typedef_type(...);
        
      };
      namespace utility {

        template<typename arg__function_pointer,
                 typename arg__tuple__of__argument_types>
        struct is_callable;

        template<typename R, typename... Ts, typename... argument_types>
        struct is_callable<R(*)(Ts...), std::tuple<argument_types...>>
          : std::is_same<Y, decltype(has_member_typedef_type<std::result_of<std::remove_pointer_t<R(*)(Ts...)>&(argument_types...)>>(0))> {
        };
        
      };
    };
  };
};
